from collections import defaultdict
from itertools import combinations
from typing import Dict
from typing import Optional
from typing import Sequence

import numpy as np
import pandas as pd
import spacy
import textacy
from cytoolz.itertoolz import concat
from cytoolz.itertoolz import sliding_window
from sklearn.metrics import pairwise
from textacy.representations import matrix_utils
from textacy.representations.vectorizers import Vectorizer


def vector_cosine_similarity(docs: Sequence[spacy.tokens.Doc]) -> np.ndarray:
    """
    Get the pairwise cosine similarity between each
    document in docs.
    """
    vectors = np.vstack([doc.vector for doc in docs])
    return pairwise.cosine_similarity(vectors)


class Conversation:
    """
    The Conversation object stores a table of conversation
    data and processes it to get it ready for analysis.

    Args:
        data: A dataframe where each row represents one turn in the
            conversation.
        text_column: The name of the column in data which holds
            the actual text of the conversation
        speaker_column: The name of the column in data which
            identifies the speaker for that turn
        id_column: An optional column name with a unique ID for
            each turn. If not provided, a default "text_id" column
            will be created.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        text_column: str = "text",
        speaker_column: str = "speaker",
        id_column: Optional[str] = None,
        language_model: str = "en_core_web_sm",
    ):
        # Set up data
        self.data = data.copy()
        if id_column is None:
            id_column = "text_id"
            self.data["text_id"] = range(len(self.data))
        self.data = self.data.rename(
            columns={
                text_column: "text",
                speaker_column: "speaker",
                id_column: "text_id",
            }
        )
        self.data.set_index("text_id", drop=False, inplace=True)
        # Apply NLP
        self.nlp = spacy.load(language_model)
        self.data["spacy_doc"] = self.data["text"].apply(self.nlp)

    def __str__(self):
        return "Conversation object:\n" + str(self.data.head())

    @property
    def n_speakers(self) -> int:
        """
        The number of speakers in the conversation.
        """
        return self.data["speaker"].nunique()

    def get_speaker_names(self) -> list:
        """
        Get a list of the speakers in the conversation.
        """
        return self.data["speaker"].unique().tolist()

    def vector_similarity_matrix(self) -> pd.DataFrame:
        """
        Calculate the similarity between each turn of the conversation
        based on spacy word vectors.
        """
        matrix = vector_cosine_similarity(self.data["spacy_doc"])
        return pd.DataFrame(matrix, index=self.data.index, columns=self.data.index)


class BaseSimilarityModel:
    """Common code for similarity models, handles creation of corpus etc."""

    def __init__(
        self,
        conversation: Conversation,
        n_top_terms: int = 20,
        top_terms: Optional[Sequence[str]] = None,
        **kwargs
    ):
        self.conversation = conversation
        self.corpus = self._create_corpus()
        # Start counting terms
        self.binary_vectorizer = Vectorizer(
            tf_type="binary",
            idf_type=None,
            dl_type=None,
        )
        self.doc_term_matrix = self.binary_vectorizer.fit_transform(
            self._get_filtered_corpus()
        )

        if top_terms is None:
            # Get top terms based on number of documents they appear in
            self.top_terms = self._get_top_terms(n_top_terms)
        else:
            # Use the user-specified terms list
            self.top_terms = top_terms

    def _create_corpus(self) -> textacy.Corpus:
        return textacy.Corpus(
            lang=self.conversation.nlp, data=self.conversation.data["text"]
        )

    @staticmethod
    def _filter_tokens(doc: spacy.tokens.Doc):
        """
        Filter out stopwords, punctuation and spaces.
        Return a generator that yields tokens, converted
        to lowercase
        """
        for token in doc:
            if not token.is_stop and not token.is_punct and not token.is_space:
                yield token.lower_

    def _get_filtered_corpus(self):
        return (self._filter_tokens(doc) for doc in self.corpus)

    def _get_top_terms(self, n: int):
        # TODO: should this work based on raw frequencies, rather than the binary
        #   per-document count?
        term_counts = matrix_utils.get_term_freqs(self.doc_term_matrix)
        top_n_indices = np.flip(np.argsort(term_counts))[:n]
        top_n_terms = [
            self.binary_vectorizer.terms_list[index] for index in top_n_indices
        ]
        return top_n_terms


class VectorSimilarityModel(BaseSimilarityModel):
    """
    Generate similarity scores from word vectors.
    Similar to the algorithm from https://doi.org/10/b49pvx
    but instead of concept vectors based on local co-occurrences
    we use the word vectors from a spacy language model.
    """

    def __init__(
        self,
        conversation: Conversation,
        n_top_terms: int = 20,
        top_terms: Optional[Sequence[str]] = None,
    ):
        super().__init__(
            conversation=conversation, n_top_terms=n_top_terms, top_terms=top_terms
        )

    def get_term_similarity_matrix(self) -> pd.DataFrame:
        """
        Calculate cosine similarity between the word vectors for each
        term in the corpus.
        """
        vectors = np.vstack(
            [
                self.conversation.nlp.vocab.get_vector(term)
                for term in self.binary_vectorizer.terms_list
            ]
        )
        similarity_matrix = pd.DataFrame(
            data=pairwise.cosine_similarity(vectors),
            index=self.binary_vectorizer.terms_list,
            columns=self.binary_vectorizer.terms_list,
        )
        return similarity_matrix

    def get_feature_vectors(self) -> pd.DataFrame:
        """
        Get the feature vectors for each document.

        Returns:
            A dataframe where the rows are the top terms in the corpus,
            and the columns are documents. Values are the feature value for
            that term's concept in each document.
        """
        term_doc_df = pd.DataFrame(
            self.doc_term_matrix.T.todense(),
            index=self.binary_vectorizer.terms_list,
            columns=self.conversation.data["text_id"],
        )
        term_similarity_matrix = self.get_term_similarity_matrix()
        key_term_similarity = term_similarity_matrix.loc[self.top_terms, :]
        return key_term_similarity @ term_doc_df

    def get_conversation_similarity(self) -> pd.DataFrame:
        feature_vectors = self.get_feature_vectors()
        doc_doc_cosine = pd.DataFrame(
            pairwise.cosine_similarity(feature_vectors.T),
            index=self.conversation.data["text_id"],
            columns=self.conversation.data["text_id"],
        )
        return doc_doc_cosine


class ConceptSimilarityModel(BaseSimilarityModel):
    """
    Generate similarity scores from local co-occurrences,
    using the algorithm from https://doi.org/10/b49pvx
    """

    def __init__(
        self,
        conversation: Conversation,
        n_top_terms: int = 20,
        top_terms: Optional[Sequence[str]] = None,
        sentence_window_size: int = 3,
    ):
        super().__init__(
            conversation=conversation, n_top_terms=n_top_terms, top_terms=top_terms
        )
        self.sentence_window_size = sentence_window_size

    @staticmethod
    def _get_sentence_windows(doc: spacy.tokens.Doc, window_size: int):
        """
        We can use cytoolz.itertoolz.sliding_window
        to generate sentence windows, but it returns
        nothing if the number of sentences is less
        than the window size. Here we just return
        the sentences for documents less than the window size.
        """
        n_sentences = sum(1 for sentence in doc.sents)
        if n_sentences < window_size:
            return [doc.sents]
        else:
            return sliding_window(window_size, (sentence for sentence in doc.sents))

    def get_cooccurrence_counts(self) -> dict:
        """
        Calculate the occurrence count of each term across all sentence
        windows, and the co-occurrence of each pair of terms

        Returns:
            A dictionary with "occurrence": Series of overall occurrence counts,
            "cooccurrence": DataFrame of term-term cooccurrence counts,
            "total_windows":
        """
        pair_counts = defaultdict(int)
        occurrence_counts = defaultdict(int)

        total_windows = 0
        for doc in self.corpus:
            windows = self._get_sentence_windows(
                doc, window_size=self.sentence_window_size
            )
            for window in windows:
                total_windows += 1
                # We only need the unique tokens so use a set
                current_tokens = set(
                    concat(self._filter_tokens(sentence) for sentence in window)
                )
                for token in current_tokens:
                    occurrence_counts[token] += 1

                pairs = combinations(sorted(current_tokens), 2)
                for pair in pairs:
                    pair_counts[pair] += 1

        # Create series and dataframe to store occurrence counts
        #   and cooccurrence counts
        occurrence_series = pd.Series(
            occurrence_counts,
            index=self.binary_vectorizer.terms_list,
            dtype=pd.Int64Dtype(),
        )
        cooccurrence_matrix = pd.DataFrame(
            0,
            index=self.binary_vectorizer.terms_list,
            columns=self.binary_vectorizer.terms_list,
            dtype=pd.Int64Dtype(),
        )
        for (term1, term2), count in pair_counts.items():
            cooccurrence_matrix.loc[term1, term2] = count
            cooccurrence_matrix.loc[term2, term1] = count

        return {
            "occurrence": occurrence_series,
            "cooccurrence": cooccurrence_matrix,
            "total_windows": total_windows,
        }

    @staticmethod
    def _get_contingency_counts(
        total_windows: int, occurrence: pd.Series, cooccurrence: pd.DataFrame
    ) -> Dict[tuple, pd.DataFrame]:
        """
        When calculating association statistics it helps to have the contingency
        counts between a term i and another term j.

        Returns:
            A dictionary mapping from 2-tuples ('i', 'j'), ('i', 'not_j'), etc. to a
            DataFrame of the counts of each.
        """
        return {
            ("i", "j"): cooccurrence,
            ("not_i", "not_j"): (
                total_windows
                - (
                    (-1 * cooccurrence)
                    .add(occurrence, axis="rows")
                    .add(occurrence, axis="columns")
                )
            ),
            ("i", "not_j"): (-1 * cooccurrence).add(occurrence, axis="rows"),
            ("not_i", "j"): (-1 * cooccurrence).add(occurrence, axis="columns"),
        }

    @classmethod
    def get_term_similarity_matrix(
        cls, total_windows: int, occurrence: pd.Series, cooccurrence: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Calculate the similarity score S(t_i, t_j) between each pair of terms t_i.

        See Angus (2012): https://doi.org/10/b49pvx
        for details of how to calculate these scores.
        """
        contingency_counts = cls._get_contingency_counts(
            total_windows=total_windows,
            occurrence=occurrence,
            cooccurrence=cooccurrence,
        )
        # Convert to probabilities
        contingency_probs = {
            k: v / total_windows for k, v in contingency_counts.items()
        }
        # Fixes for zero counts/terms only appearing in one context
        contingency_probs[("i", "not_j")] = contingency_probs[("i", "not_j")].where(
            ~cooccurrence.eq(occurrence, axis="rows"), 1
        )
        contingency_probs[("not_i", "j")] = contingency_probs[("not_i", "j")].where(
            ~cooccurrence.eq(occurrence, axis="columns"), 1
        )

        similarity_matrix = (
            contingency_probs[("i", "j")] * contingency_probs[("not_i", "not_j")]
        ) / (contingency_probs[("not_i", "j")] * contingency_probs[("i", "not_j")])
        return similarity_matrix

    def get_concept_vectors(self, term_similarity_matrix: pd.DataFrame) -> pd.DataFrame:
        """
        Get the feature vectors for each document.

        Returns:
            A dataframe where the rows are the top terms in the corpus,
            and the columns are documents. Values are the feature value for
            that term's concept in each document.
        """
        term_doc_df = pd.DataFrame(
            self.doc_term_matrix.T.todense(),
            index=self.binary_vectorizer.terms_list,
            columns=self.conversation.data["text_id"],
        )
        key_term_similarity = term_similarity_matrix.loc[self.top_terms, :]
        return key_term_similarity @ term_doc_df

    def get_conversation_similarity(self):
        counts = self.get_cooccurrence_counts()
        term_similarity_matrix = self.get_term_similarity_matrix(**counts)
        concept_vectors = self.get_concept_vectors(term_similarity_matrix)

        doc_doc_cosine = pd.DataFrame(
            pairwise.cosine_similarity(concept_vectors.T),
            index=self.conversation.data["text_id"],
            columns=self.conversation.data["text_id"],
        )
        return doc_doc_cosine



from typing import Callable
from typing import Optional

import numpy as np
import pandas as pd
from bokeh import layouts
from bokeh import models
from bokeh import palettes
from bokeh.plotting import figure
from bokeh.plotting import show
from bokeh.themes import Theme
from bokeh.transform import factor_cmap

from .conversation import Conversation

BLANK_PLOT_THEME = Theme(
    json={
        "attrs": {
            "Figure": {
                "background_fill_color": "#FFFFFF",
                "border_fill_color": "#FFFFFF",
                "outline_line_color": "#000000",
            },
            "Axis": {
                "axis_line_color": None,
            },
            "Grid": {"grid_line_color": None},
            "Title": {"text_color": "black"},
        }
    }
)

WORD_WRAP_CELL_TEMPLATE = """
<span style="word-wrap: break-word; max-width: 600px;
      display: inline-block; white-space: normal;">
<%= value %>
</span>
"""


PLOT_HELP_TEXT = """
<h3>Similarity plot</h3>

<p>
Click an item on the diagonal to view it in the table below.
Click anywhere on the background to deselect it.
</p>
"""


def _get_word_wrap_formatter():
    """
    We can't store this formatter as a constant or it causes
    issues in Jupyter notebooks to do with reusing
    models
    """
    return models.widgets.tables.HTMLTemplateFormatter(template=WORD_WRAP_CELL_TEMPLATE)


class ConversationPlot:
    """
    Create an interactive conversation plot from a Conversation object.

    Args:
        conversation: A Conversation object.
        similarity_matrix: An optional matrix of similarity scores between each
            turn in the conversation. If this isn't provided, spacy's default
            similarity calculation for documents (cosine similarity between
            average word embeddings) will be used.
            The index and column names for the matrix must match the text_id from
            the conversation.
        options: A dictionary of options for the visual style of the plot. See
            ConversationPlot.DEFAULT_OPTIONS for the available options.
    """

    DEFAULT_OPTIONS = {"width": 800, "height": 800, "tile_padding": 1}

    def __init__(
        self,
        conversation: Conversation,
        similarity_matrix: Optional[pd.DataFrame] = None,
        options: Optional[dict] = None,
    ):
        self.options = self.DEFAULT_OPTIONS.copy()
        if options is not None:
            self.options.update(options)

        long_data = self._get_long_data(conversation, similarity_matrix)
        self.n_speakers = conversation.n_speakers
        self.speakers = conversation.get_speaker_names()
        self.diagonal_datasource = self._get_diagonal_datasource(
            conversation, long_data=long_data
        )
        self.similarity_datasource = self._get_similarity_datasource(
            long_data=long_data
        )

    def _get_diagonal_datasource(
        self, conversation: Conversation, long_data: pd.DataFrame
    ) -> models.ColumnDataSource:
        """
        Get the data for the texts on the diagonal, as a bokeh ColumnDataSource
        """
        text_data = long_data.query("x_index == y_index").copy()
        text_data["text"] = text_data["x_index"].map(conversation.data["text"])
        return models.ColumnDataSource(text_data)

    def _get_similarity_datasource(
        self, long_data: pd.DataFrame
    ) -> models.ColumnDataSource:
        """
        Get the data for the similarity cells in the lower triangle,
        including calculating coordinates for the triangles which
        will be displayed.
        """
        similarity_data = long_data.query("x_index < y_index").copy()
        half_width = similarity_data["width"] / 2
        half_height = similarity_data["height"] / 2
        triangle_coords = pd.DataFrame(index=similarity_data.index)
        triangle_coords["left"] = similarity_data["x_position"] - half_width
        triangle_coords["right"] = similarity_data["x_position"] + half_width
        # We flip the y-axis for the plot, so subtract to get to the top
        triangle_coords["top"] = similarity_data["y_position"] - half_height
        triangle_coords["bottom"] = similarity_data["y_position"] + half_height

        # Having some issues converting nested pandas cols to ColumnDataSource,
        #   convert to dict before we try to convert to ColumnDataSource
        similarity_dict = similarity_data[
            ["x_index", "y_index", "x_speaker", "y_speaker", "similarity"]
        ].to_dict(orient="list")

        # Get upper triangle coordinates: bottom left, top left, top right
        # Use df.values.tolist() to quickly create nested lists from df rows
        # bokeh's multi_polygon plot function actually needs triply-nested lists,
        # because it allows for plotting polygons with multiple holes
        similarity_dict["upper_triangle_x"] = [
            [[row]]
            for row in triangle_coords[["left", "left", "right"]].values.tolist()
        ]
        similarity_dict["upper_triangle_y"] = [
            [[row]] for row in triangle_coords[["bottom", "top", "top"]].values.tolist()
        ]

        # Get lower triangle coordinates: bottom left, top right, bottom right
        similarity_dict["lower_triangle_x"] = [
            [[row]]
            for row in triangle_coords[["left", "right", "right"]].values.tolist()
        ]
        similarity_dict["lower_triangle_y"] = [
            [[row]]
            for row in triangle_coords[["bottom", "top", "bottom"]].values.tolist()
        ]

        return models.ColumnDataSource(similarity_dict)

    def _get_long_data(
        self,
        conversation: Conversation,
        similarity_matrix: Optional[pd.DataFrame] = None,
    ):
        """
        Convert the similarity matrix to long form for plotting.
        """
        if similarity_matrix is None:
            similarity_matrix = pd.DataFrame(conversation.vector_similarity_matrix())
        # Info about each text, for merging with similarity data
        text_info = pd.DataFrame(
            {
                "n_words": conversation.data["spacy_doc"].map(len),
                "speaker": conversation.data["speaker"],
            }
        )
        # Log-scale number of words for rectangle sizes
        text_info["size"] = np.log(text_info["n_words"])
        # Position is half the width of the current tile plus half the width,
        #   then we add tile_padding between each
        text_info["position"] = (
            (text_info["size"] / 2) + (text_info["size"].shift(1, fill_value=0) / 2)
        ).cumsum() + (np.arange(len(text_info)) * self.options["tile_padding"])

        similarity = (
            similarity_matrix.melt(
                ignore_index=False, var_name="y_index", value_name="similarity"
            )
            # Copy index into x_index column
            .assign(x_index=lambda df: df.index)
            # Don't need the upper triangle
            .query("x_index <= y_index")
        )
        # Copy x index
        similarity["x_index"] = similarity.index
        # Get x properties
        similarity["width"] = similarity["x_index"].map(text_info["size"])
        similarity["x_position"] = similarity["x_index"].map(text_info["position"])
        similarity["x_speaker"] = similarity["x_index"].map(text_info["speaker"])
        similarity["n_words"] = similarity["x_index"].map(text_info["n_words"])
        # Get y properties
        similarity["height"] = similarity["y_index"].map(text_info["size"])
        similarity["y_position"] = similarity["y_index"].map(text_info["position"])
        similarity["y_speaker"] = similarity["y_index"].map(text_info["speaker"])

        return similarity

    def _create_text_table(self):
        """
        Create the text table widget that shows information for clicked/selected
        tiles
        """
        columns = [
            models.TableColumn(field="x_speaker", title="speaker"),
            models.TableColumn(
                field="text", title="text", formatter=_get_word_wrap_formatter()
            ),
        ]
        # Start off showing nothing
        empty_view = models.CDSView(
            source=self.diagonal_datasource, filters=[models.IndexFilter([])]
        )
        table = models.DataTable(
            source=self.diagonal_datasource,
            columns=columns,
            autosize_mode="fit_columns",
            sizing_mode="stretch_width",
            row_height=100,
            view=empty_view,
        )
        return table

    def _get_categorical_palette(
        self, n_colours: int, palette_name: str = "Category10"
    ):
        """
        Return a palette of n_colours
        """
        # bokeh doesn't have 2-colour palettes, just use the 3-colour
        palette_n = max(n_colours, 3)
        return palettes.d3[palette_name][palette_n][:n_colours]

    def show_plot(self):
        """
        Show the interactive plot as a Jupyter notebook output.
        Requires bokeh.io.output_notebook() to have been run.
        """
        plot_func = self.create_plot_function()
        try:
            show(plot_func)
        except AssertionError:
            raise RuntimeError(
                "No bokeh output detected. Make sure you run "
                "bokeh.io.output_notebook() at the top of your notebook"
                " to enable output"
            ) from None

    def create_plot_function(self) -> Callable:
        """
        In order to have interactivity controlled by Python callbacks,
        bokeh requires us to wrap the plot creation in a function that
        accepts a "document" argument.
        """

        def _add_plot_tools(plot):
            click_to_select = models.TapTool()
            hover_similarity = models.HoverTool(
                names=["similarity_upper", "similarity_lower"],
                tooltips=[("Similarity", "@similarity")],
            )
            hover_text = models.HoverTool(
                names=["text_tiles"], tooltips=[("Text", "@text")]
            )
            plot.add_tools(hover_similarity, hover_text, click_to_select)
            plot.toolbar.active_tap = click_to_select

        def plot_func(doc):
            doc.theme = BLANK_PLOT_THEME

            # Table for viewing texts
            text_table = self._create_text_table()

            def _set_table_filter(attr, old, new):
                text_table.view = models.CDSView(
                    filters=[models.IndexFilter(new)], source=self.diagonal_datasource
                )

            self.diagonal_datasource.selected.on_change("indices", _set_table_filter)

            # Main plot ############
            plot = figure(
                width=self.options["width"],
                height=self.options["height"],
                aspect_scale=1.0,
            )

            # Set up speaker colours
            speaker_colours = self._get_categorical_palette(self.n_speakers)
            speaker_cmap = factor_cmap(
                "x_speaker", palette=speaker_colours, factors=self.speakers
            )
            other_speaker_cmap = factor_cmap(
                "y_speaker", palette=speaker_colours, factors=self.speakers
            )
            # Plot similarity tiles
            plot.multi_polygons(
                xs="upper_triangle_x",
                ys="upper_triangle_y",
                alpha="similarity",
                color=speaker_cmap,
                source=self.similarity_datasource,
                line_width=0,
                name="similarity_upper",
            )
            plot.multi_polygons(
                xs="lower_triangle_x",
                ys="lower_triangle_y",
                alpha="similarity",
                color=other_speaker_cmap,
                source=self.similarity_datasource,
                line_width=0,
                name="similarity_lower",
            )

            # Plot diagonal/text tiles
            plot.rect(
                x="x_position",
                y="y_position",
                color=speaker_cmap,
                source=self.diagonal_datasource,
                name="text_tiles",
                legend_field="x_speaker",
            )

            # Options/style
            plot.y_range.flipped = True
            plot.axis.major_tick_line_color = None
            plot.axis.minor_tick_line_color = None
            plot.axis.major_label_text_color = None
            plot.legend.title = "Speaker"

            # Interactivity/plot tools
            _add_plot_tools(plot)

            # Create layout
            doc.add_root(
                layouts.column(
                    layouts.row(models.Div(text=PLOT_HELP_TEXT, width=400)),
                    layouts.row(plot),
                    layouts.row(models.Div(text="<h3>Selected text</h3>")),
                    layouts.row(text_table),
                    sizing_mode="stretch_width",
                )
            )

        return plot_func



import math
import re
import uuid
from typing import Union

import ipywidgets
import pandas as pd
import spacy
from IPython.display import display
from textacy.extract import keyword_in_context

SEARCH_CSS_TEMPLATE = """
<style>
#{element_id} table {{
  background-color: #f9f9f9;
}}

#{element_id} tbody tr:nth-child(odd) {{
  background-color: #f9f9f9;
}}

#{element_id} tbody tr:nth-child(even) {{
  background-color: #dddddd;
}}

#{element_id} .atap.search_highlight {{
    background-color: #99CC99;
    padding-left: 0px;
    padding-right: 0px;
}}

#{element_id} .atap.context {{
    color: #444;
    white-space: pre;
}}
#{element_id} .atap.context.context_left {{
    text-align: right;
    padding-right: 0px;
}}
#{element_id} .atap.context.context_right {{
    text-align: left;
    padding-left: 0px;
}}
#{element_id}.search_table th {{
    text-align: left;
}}

#{element_id}.search_table td {{
    border: none;
}}

#{element_id} td.text_id {{
    border-right: 1px solid #bdbdbd;
}}

#{element_id} .atap.regex_error {{
    background-color: #CC9999;
}}
</style>
"""

SEARCH_TABLE_TEMPLATE = """
{css}
<table id="{element_id}" class="atap search_table">
    <thead>
        <tr>
            <th>Document ID</th>
            <th>Result</th>
            <th/>
            <th/>
        </tr>
    </thead>
    <tbody>
        {table_rows}
    </body>
</table>
{n_results} results. A maximum of {max_results} are displayed.
"""

SEARCH_ROW_TEMPLATE = """
<tr>
    <td class="atap text_id">{text_id}</td>
    <td class="atap context context_left">{left_context}</td>
    <td class="atap search_highlight">{match}</td>
    <td class="atap context context_right">{right_context}</td>
</tr>
"""

REGEX_ERROR_TEMPLATE = """
{css}
<div id="{element_id}">
<span class="atap regex_error">
Incomplete or incorrect regular expression: {error.msg}
</span>
</div>
"""


class NoResultsError(Exception):
    pass


def prepare_text_df(
    df: pd.DataFrame,
    text_column: str = "text",
    id_column: str = None,
    language_model: Union[str, spacy.language.Language] = "en_core_web_sm",
) -> pd.DataFrame:
    """
    Our text processing functions expect a dataframe with
    a few required columns:

    - "text": containing the raw text data
    - "text_id": a unique ID for each text
    - "spacy_doc": the text, processed into a spacy Doc

    prepare_text_df() takes an input DataFrame, df, and creates/renames
    columns as necessary to match this format.

    Args:
        df: Input dataframe containing multiple texts.
        text_column: The current name of the text column in df
        id_column: The current column name of the unique identifier for each text
            in df. If not given, numeric IDs will be generated for each text.
        language_model: The name of a spacy model like "en_core_web_sm", or a
            spacy language model instance.
    """
    output = df.copy()
    if id_column is None:
        output["text_id"] = pd.Series(range(output.shape[0]), dtype=pd.Int64Dtype)
        id_column = "text_id"
    output = output.rename(columns={text_column: "text", id_column: "text_id"})
    output = output.set_index("text_id", drop=False)

    if isinstance(language_model, str):
        language_model = spacy.load(language_model)
    output["spacy_doc"] = output["text"].map(language_model)
    return output


class ConcordanceWidget:
    """
    Interactive widget for searching and displaying concordance
    results
    """

    def __init__(self, df: pd.DataFrame, results_per_page: int = 20):
        """
        Args:
            df: DataFrame containing texts, as returned by prepare_text_df()
            results_per_page: How many search results to show at a time
        """
        self.data = df
        self.results_per_page = results_per_page
        self.search_table = ConcordanceTable(df=self.data)

    def show(self):
        """
        Display the interactive widget
        """

        def display_results(page: int, **kwargs):
            if not kwargs["keyword"]:
                return None
            for attr, value in kwargs.items():
                setattr(self.search_table, attr, value)

            try:
                # This will either be results or a regex error message
                html = self.search_table._get_html(page=page)
                if self.search_table.is_regex_valid():
                    results = self.search_table._get_results()
                    # Need at least one page
                    n_pages = max(
                        self.search_table._get_total_pages(n_results=len(results)), 1
                    )
                    page_input.max = n_pages
            except NoResultsError:
                n_pages = 1

            display(ipywidgets.HTML(html))

        keyword_input = ipywidgets.Text(description="Keyword(s):")
        regex_toggle_input = ipywidgets.Checkbox(
            value=False,
            description="Enable regular expressions",
            disabled=False,
            style={"description_width": "initial"},
        )
        ignore_case_input = ipywidgets.Checkbox(
            value=True, description="Ignore case", disabled=False
        )
        whole_word_input = ipywidgets.Checkbox(
            value=False,
            description="Match whole words",
            disabled=False,
            style={"description_width": "initial"},
        )
        page_input = ipywidgets.BoundedIntText(
            value=1,
            min=1,
            max=1,
            step=1,
            description="Page:",
        )
        window_width_input = ipywidgets.BoundedIntText(
            value=50,
            min=10,
            step=1,
            description="Window size (characters):",
            style={"description_width": "initial"},
        )
        sort_input = ipywidgets.Dropdown(
            options=["text_id", "left_context", "right_context"],
            value="text_id",
            description="Sort by:",
        )
        output = ipywidgets.interactive_output(
            display_results,
            {
                "keyword": keyword_input,
                "regex": regex_toggle_input,
                "ignore_case": ignore_case_input,
                "whole_word": whole_word_input,
                "page": page_input,
                "window_width": window_width_input,
                "sort": sort_input,
            },
        )
        # Excel export
        filename_input = ipywidgets.Text(
            value="",
            placeholder="Enter filename",
            description="Filename (without.xlsx extension)",
            style={"description_width": "initial"},
            disabled=False,
        )
        export_button = ipywidgets.Button(
            description="Export to Excel",
            disabled=False,
            button_style="success",
            tooltip="The excel file will be saved to the same location as the "
            "notebook on the server. Use the Jupyter Lab sidebar to access "
            "and download it.",
        )

        def _export(widget):
            filename = filename_input.value
            # TODO: raise an error here?
            if not filename:
                return

            if not filename.endswith(".xlsx"):
                filename = filename + ".xlsx"

            self.search_table.to_excel(filename)

        export_button.on_click(_export)
        export_controls = ipywidgets.HBox([filename_input, export_button])

        # Set up layout of widgets
        checkboxes = ipywidgets.HBox(
            [regex_toggle_input, ignore_case_input, whole_word_input]
        )
        checkboxes.layout.justify_content = "flex-start"
        number_inputs = ipywidgets.HBox([page_input, window_width_input])
        number_inputs.layout.justify_content = "flex-start"
        return ipywidgets.VBox(
            [
                keyword_input,
                checkboxes,
                number_inputs,
                sort_input,
                export_controls,
                output,
            ]
        )


class ConcordanceTable:
    """
    Search for matches in a text (using plain-text or regular expressions),
    and display them in a HTML table.

    The arguments given to the constructor set up the table with
    initial data and search settings. In order
    to allow interactive use, these attributes can be changed
    after initialization and the updated values will be used
    next time a search is triggered.

    Args:
        df: DataFrame of texts, as returned by prepare_text_df().
        keyword: Word, phrase or regular expression to search for.
        regex: Use regular expression matching?
        ignore_case: If False, searches are case-sensitive.
        whole_word: Only return matches where the search term matches
           whole words (with space or punctuation either side)
        results_per_page: Number of results to show at a time.
        window_width: Number of characters to show either side of a match.
        sort: Sort by 'text_id', 'left_context' or 'right_context'.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        keyword: str = "",
        regex: bool = False,
        ignore_case: bool = True,
        whole_word: bool = False,
        results_per_page: int = 20,
        window_width: int = 50,
        sort: str = "text_id",
    ):
        self.df = df
        self.keyword = keyword
        self.regex = regex
        self.ignore_case = ignore_case
        self.whole_word = whole_word
        self.results_per_page = results_per_page
        self.window_width = window_width
        self.sort = sort
        self.element_id = "atap_" + str(uuid.uuid4())
        self.css = SEARCH_CSS_TEMPLATE.format(element_id=self.element_id)

    def _repr_mimebundle_(self, include, exclude):
        """
        Define _repr_mimebundle_() so the default behaviour in Jupyter
        notebooks is to view the HTML table.
        """
        return {"text/html": self._get_html()}

    def _get_search_regex(self) -> re.Pattern:
        """
        Compile a regex we can pass to keyword_in_context.
        Will be escaped to only return literal matches if
        self.regex = False.
        Will raise a re.error exception if the regex is invalid
        (useful for showing feedback to the user).

        Only call this at display time because there's a decent
        chance the regex will be invalid while the user is typing,
        and we just handle the exception as part of the display logic.
        """
        if not self.regex:
            regex = re.escape(self.keyword)
        else:
            regex = self.keyword
        if self.whole_word:
            regex = f"\\b{regex}\\b"

        flags = re.IGNORECASE if self.ignore_case else 0
        return re.compile(regex, flags=flags)

    def is_regex_valid(self) -> bool:
        """
        Test if the current keyword is a valid regular expression.
        """
        try:
            _ = self._get_search_regex()
            return True
        except re.error:
            return False

    def _get_results(self) -> pd.DataFrame:
        """
        Return a Series of matches, with text_id as the index. Each element
        is a match returned by keyword_in_context(), i.e. a left_context,
        match, right_context tuple.
        """

        def _get_matches(doc):
            matches = keyword_in_context(
                doc, keyword=search_regex, window_width=self.window_width
            )
            return list(matches)

        search_regex = self._get_search_regex()
        search_results = self.df["spacy_doc"].apply(_get_matches)
        search_results = search_results.loc[search_results.map(len) > 0].explode()
        search_results.name = "match"

        if len(search_results) == 0:
            raise NoResultsError("No results found.")

        results_df = search_results.to_frame()
        # Use apply(pd.Series) to unpack nested lists into columns
        results_df[["left_context", "match", "right_context"]] = results_df[
            "match"
        ].apply(pd.Series)
        results_df.index.name = "text_id"
        results_df.reset_index(inplace=True)
        # Reorder columns
        results_df = results_df[["text_id", "left_context", "match", "right_context"]]
        return results_df

    def _get_total_pages(self, n_results: int) -> int:
        return math.ceil(n_results / self.results_per_page)

    def _get_table_html(self, search_results: pd.DataFrame, n_total: int) -> str:
        table_rows = "\n".join(
            SEARCH_ROW_TEMPLATE.format(**row).replace(r"\n", "")
            for index, row in search_results.iterrows()
        )

        html = SEARCH_TABLE_TEMPLATE.format(
            css=self.css,
            element_id=self.element_id,
            table_rows=table_rows,
            n_results=n_total,
            max_results=self.results_per_page,
        )
        return html

    def _get_error_html(self, error) -> str:
        return REGEX_ERROR_TEMPLATE.format(
            css=self.css, element_id=self.element_id, error=error
        )

    def _get_html(self, page: int = 1):
        regex_valid = self.is_regex_valid()
        if not regex_valid:
            try:
                self._get_search_regex()
            except re.error as error:
                return self._get_error_html(error)

        start_index = (page - 1) * self.results_per_page
        end_index = (page * self.results_per_page) - 1
        try:
            results = self._get_results()
            n_total = len(results)
            results = results.iloc[start_index:end_index]
        except NoResultsError:
            return "No results found. Try a different search term"

        if self.sort == "text_id":
            results = results.sort_values(by="text_id")
        elif self.sort in ("left_context", "right_context"):
            results = self.sort_by_context(results, context=self.sort)
        else:
            raise ValueError(
                f"Invalid sort value {self.sort}: should be 'text_id',"
                " 'left_context' or 'right_context'"
            )

        return self._get_table_html(results, n_total=n_total)

    def to_dataframe(self) -> pd.DataFrame:
        """
        Return a dataframe returning all the matches for the current
        keyword.
        """
        return self._get_results()

    def to_excel(self, filename: str, max_col_width: int = 100):
        """
        Export matches for the current keyword to Excel.

        Args:
            filename: Path to write the Excel file to.
            max_col_width: Maximum width of Excel columns. We try to automatically
                set the widths of columns to fit their content, but only
                up to this maximum.
        """

        def _set_col_widths():
            for column in df:
                col_width = max(df[column].astype(str).map(len).max(), len(column))
                col_width = min(col_width, max_col_width)
                col_index = df.columns.get_loc(column)
                writer.sheets["results"].set_column(col_index, col_index, col_width)

        df = self.to_dataframe()
        # Need xlsxwriter for (relatively) easy setting of column widths
        writer = pd.ExcelWriter(filename, engine="xlsxwriter")
        df.to_excel(writer, index=False, sheet_name="results")
        # Formatting
        right_align = writer.book.add_format()
        right_align.set_align("right")
        left_align = writer.book.add_format()
        left_align.set_align("left")
        writer.sheets["results"].set_column("B:B", None, right_align)
        writer.sheets["results"].set_column("D:D", None, left_align)
        _set_col_widths()

        writer.save()

    @staticmethod
    def sort_by_context(results: pd.DataFrame, context: str = "left_context"):
        """
        Sort concordance results by either the left context or right
        context. For left context, this means sorting by the preceding
        word, then the word before that, etc.
        """

        def get_reversed_words(s: pd.Series):
            return s.str.strip().str.lower().str.split(r"\s").map(lambda x: x[::-1])

        if context == "left_context":
            return results.sort_values(by=context, key=get_reversed_words)
        elif context == "right_context":
            return results.sort_values(
                by=context, key=lambda x: x.str.strip().str.lower()
            )
        else:
            raise ValueError(
                "Invalid context option: should be 'left_context' or 'right_context'"
            )
